package edu.cmu.ri.createlab.serial;

import java.util.Arrays;
import edu.cmu.ri.createlab.util.commandexecution.CommandResponse;

/**
 * @author Chris Bartley (bartley@cmu.edu)
 */
public final class SerialDeviceCommandResponse implements CommandResponse
   {
   private static final String EOL = System.getProperty("line.separator", "\n");

   private final boolean success;
   private final byte[] data;

   /** Creates a new SerialDeviceCommandResponse having the given success and <code>null</code> data. */
   public SerialDeviceCommandResponse(final boolean success)
      {
      this(success, null);
      }

   /**
    * Creates a new SerialDeviceCommandResponse having the given success and data.  This constructor does NOT create a
    * copy of the given byte array, so the caller must be careful to not mutate the array after calling this constructor.
    */
   @SuppressWarnings({"AssignmentToCollectionOrArrayFieldFromParameter"})
   public SerialDeviceCommandResponse(final boolean success, final byte[] data)
      {
      this.success = success;
      this.data = data;
      }

   /**
    * Creates a new SerialDeviceCommandResponse using the given data.  The status code is <code>true</code> if the given
    * data array is non-null; <code>false</code> otherwise.  This constructor does NOT create a copy of the given byte
    * array, so the caller must be careful to not mutate the array after calling this constructor.
    */
   public SerialDeviceCommandResponse(final byte[] data)
      {
      this(data != null, data);
      }

   /**
    * Returns <code>true</code> if the command was successful; <code>false</code> otherwise.  It is up to each command
    * to define what success means.
    */
   @Override
   public boolean wasSuccessful()
      {
      return success;
      }

   /**
    * Returns the data as an array of bytes.  May return null.  Note that the underlying array is returned, so changes
    * to the array will change the instance.
    */
   @SuppressWarnings({"ReturnOfCollectionOrArrayField"})
   @Override
   public byte[] getData()
      {
      return data;
      }

   /**
    * Indicates whether some other object is "equal to" this one. Note that since it's possible to mutate the underlying
    * data array, users expecting consistent results from <code>equals()</code> must be careful to not modify the array.
    *
    * @see #getData()
    * @see #SerialDeviceCommandResponse(byte[])
    * @see #SerialDeviceCommandResponse(boolean, byte[])
    */
   public boolean equals(final Object o)
      {
      if (this == o)
         {
         return true;
         }
      if (o == null || getClass() != o.getClass())
         {
         return false;
         }

      final SerialDeviceCommandResponse that = (SerialDeviceCommandResponse)o;

      if (success != that.success)
         {
         return false;
         }
      if (!Arrays.equals(data, that.data))
         {
         return false;
         }

      return true;
      }

   public int hashCode()
      {
      int result = (success ? 1 : 0);
      result = 31 * result + (data != null ? Arrays.hashCode(data) : 0);
      return result;
      }

   public String toString()
      {
      final String dataSize = data == null ? "null" : data.length + " byte(s)";
      final StringBuilder s = new StringBuilder("SerialDeviceCommandResponse{" + EOL);
      s.append("   success: ").append(success).append(EOL);
      s.append("   data:    ").append(dataSize).append(EOL);
      if ((data != null) && (data.length > 0))
         {
         for (final byte b : data)
            {
            final int theByte = b & 0xff;
            s.append("      ").append(theByte).append(EOL);
            }
         }
      s.append("}").append(EOL);

      return s.toString();
      }
   }